﻿/*
 *	ID:		FLDIGI.cs
 *	Title:	Driver for FLDIGI Software
 *	Author:	T.G.Glagowski / W1TR / AFA1DI
 *	Org:	USAF MARS, ARRL, YCCC, NeWS Group, CMARA
 *	Create:	27-Aug-2013 11:28 - MSVS AppWizard: C# Class
 *	Ver:	16-Sep-2013 15:47
 *	Description:
 *		Driver for FLDIGI Digital Communication Software by W1HKJ
 *	References:
 *		http://en.wikipedia.org/wiki/XML-RPC
 *		http://xmlrpc.scripting.com/default.html
 *		http://www.w1hkj.com/FldigiHelp-3.22/xmlrpc-control.html
 *	Changes:
 *		27-Aug-2013 15:11 - API Functions for each FLDIGI Method Signature Type
 *		28-Aug-2013 09:45 - API Functions for each FLDIGI Method
 *		28-Aug-2013 10:09 - Fields and Properties for Error Code and Error Message Aded
 *		28-Aug-2013 23:37 - Argument Generation Bug Fixed
 *		09-Sep-2013 19:35 - Fault Response Structure Implemented, Reporting of Network and XML Errors
 *		11-Sep-2013 20:35 - Structure and Array Arguments and Return Values
 *		12-Sep-2013 12:32 - Simulated Query and Response from XML Files Option
 *		12-Sep-2013 13:01 - Removed All STATIC Data and Functions, Except ASCII <==> BASE64 Translation Functions
 *		12-Sep-2013 15:59 - Added Query / Response Log File Option 
 *		14-Sep-2013 10:39 - Module Separated into a C# Class Library DLL
 *		15-Sep-2013 17:45 - XML RPC Server Name Property Added
 *		15-Sep-2013 17:54 - Version Property Added
 *		16-Sep-2013 15:47 - Version Date/Time Stamp Derived from Build Date/Time
 */

//	BUG:	modem.set_bandwidth crashes FLDIGI when argument is bad range

//	BUG:	Bandwidth Family does NOT Work

// Generated by AppWizard
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

// Added Manually
using System.Xml;								// Added for XML Document
using System.Net;								// Added for TCP/IP
using System.IO;								// Added for Basic File I/O, FileInfo, File, Path, Directory


// Custom Libraries
namespace TGG_LIB.FLDIGI
{
	/////	FLDIGI ERROR CODES
	///////////////////////////////////////////////////////////////////////////////////////////

	/// <summary>Error Codes for FLDIGI Class</summary>
	public enum FLDIGI_ErrorCategory
	{
		/// <summary>OK, No Error</summary>
		OK=0,
	
		/// <summary>XML RPC Server Error Occurred</summary>
		XML_RPC_ServerError,

		/// <summary>Network Error Occurred</summary>
		NetworkError,

		/// <summary>Data Error Occurred</summary>
		DataError,

		/// <summary>XML Error Occurred</summary>
		XML_Error,

		/// <summary>Miscellaneous Error Occurred</summary>
		Error
	}

	/////	FLDIGI CLASS
	///////////////////////////////////////////////////////////////////////////////////////////

	/// <summary>Class to Operate FLDIGI</summary>
	//	http://www.w1hkj.com/FldigiHelp-3.21/xmlrpc-control.html - FLDIGI Reference
	public class FLDIGI
	{
		//	CONSTANTS
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>FLDIGI XML-RPC Server Name and Port</summary>
		private const String c_DefaultServerName = "http://localhost:7362";

		//	VARIABLE FIELDS
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>Unique Error Number</summary>
		/// <remarks>This helps trace the occurrence of an error back to source code</remarks>
		private Int32 m_ErrorNumber;

		/// <summary>Error Category</summary>
		private FLDIGI_ErrorCategory m_ErrorCategory;

		/// <summary>Error Name, Specific Error</summary>
		private String m_ErrorName;

		/// <summary>Error Message, Information Forwarded from Another Source, Such as an Exception Message or Subordinate Module</summary>
		private String m_ErrorMessage;

		/// <summary>Simulated Query Filename</summary>
		/// <remarks>When NOT Empty, Contains the Filename of a Simulated XML Format Query</remarks>
		String m_SimulatedQueryFilename;

		/// <summary>Simulated Response Filename</summary>
		/// <remarks>When NOT Empty, Contains the Filename of a Simulated XML Format Response</remarks>
		String m_SimulatedResponseFilename;

		/// <summary>Query Response Log Filename</summary>
		/// <remarks>When NOT Empty, Contains the Filename of a Query Response Log File</remarks>
		String m_QueryResponseLogFilename;

		/// <summary>FLDIGI XML-RPC Server Name and Port</summary>
		private String m_XmlRpcServerName;

		//	CONSTRUCTORS / DESTRUCTORS / INITIALIZERS
		////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>Default Constructor</summary>
		public FLDIGI()
		{
			InitializeObject();
		}
		//=========================================================================================

		/// <summary>Initialize FLDIGI Object</summary>
		private void InitializeObject()
		{
			m_ErrorNumber = 0;
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;
			m_ErrorMessage = String.Empty;
			m_ErrorName = String.Empty;
			m_SimulatedQueryFilename = String.Empty;
			m_SimulatedResponseFilename = String.Empty;
			m_QueryResponseLogFilename = String.Empty;
			m_XmlRpcServerName = c_DefaultServerName;
		}
		//=========================================================================================

		//	PROPERTIES
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>Error Code Property, Error Category</summary>
		public FLDIGI_ErrorCategory ErrorCode
		{
			get
			{
				return m_ErrorCategory;
			}
			set
			{
			}
		}
		//=========================================================================================

		/// <summary>Error Number Property, Unique Identifier</summary>
		/// <remarks>This helps trace error back to code</remarks>
		public Int32 ErrorNumber
		{
			get
			{
				return m_ErrorNumber;
			}
			set
			{
			}
		}
		//=========================================================================================

		/// <summary>Error Message Property, Information Forwarded from Another Source, Such as an Exception Message</summary>
		public String ErrorMessage
		{
			get
			{
				return m_ErrorMessage;
			}
			set
			{
			}
		}
		//=========================================================================================

		/// <summary>Error Name Property, Specific Error</summary>
		public String ErrorName
		{
			get
			{
				return m_ErrorName;
			}
			set
			{
			}
		}
		//=========================================================================================

		/// <summary>Simulated Response Filename Property</summary>
		/// <remarks>When NOT Empty String, Use File as Query for the XML-RPC Server, Used for Debugging</remarks>
		public String SimulatedQueryFilename
		{
			get
			{
				return m_SimulatedQueryFilename;
			}
			set
			{
				m_SimulatedQueryFilename = value;
			}
		}
		//=========================================================================================

		/// <summary>Simulated Response Filename Property</summary>
		/// <remarks>When NOT Empty String, Use File as Response instead of Querying XML-RPC Server, Used for Debugging</remarks>
		public String SimulatedResponseFilename
		{
			get
			{
				return m_SimulatedResponseFilename;
			}
			set
			{
				m_SimulatedResponseFilename = value;
			}
		}
		//=========================================================================================

		/// <summary>Query Response Log Filename Property</summary>
		/// <remarks>When NOT Empty String, Use File to Log Queries and Responses</remarks>
		public String QueryResponseLogFilename
		{
			get
			{
				return m_QueryResponseLogFilename;
			}
			set
			{
				m_QueryResponseLogFilename = value;
			}
		}
		//=========================================================================================

		/// <summary>XML-RPC Server Name</summary>
		public String XmlRpcServerName
		{
			get
			{
				return m_XmlRpcServerName;
			}
			set
			{
				m_XmlRpcServerName = value;
			}
		}
		//=========================================================================================

		/// <summary>DLL Version</summary>
		/// <remarks>Version is Date/Time Stamp, Format: DD-MMM-YYYY hh:mm:ss</remarks>
		public static String Version
		{
			get
			{
				String myDllVersion;
				String myDllFilename;
				FileInfo myFileInfo;
				DateTime myDllBuildTime;

				if (File.Exists(myDllFilename = "FLDG.DLL"))
				{
					myFileInfo = new FileInfo(myDllFilename);
					myDllBuildTime = myFileInfo.LastWriteTime;
					myDllVersion = String.Format("{0}", myDllBuildTime);
				}
				else
				{
					myDllVersion = "N/A";
				}

				return myDllVersion;
			}
			set
			{
			}
		}
		//=========================================================================================

		//	FUNCTIONS
		////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>Encode ASCII String to BASE64</summary>
		/// <param name="theAsciiString">the String</param>
		/// <returns>Encoded BASE64 String</returns>
		/// http://en.wikipedia.org/wiki/Base64
		public static String Base64_Encode(
			String theAsciiString
			)
		{
			byte[] myAsciiBytes;
			String myBase64String;

			myAsciiBytes = Encoding.ASCII.GetBytes(theAsciiString);
			myBase64String = Convert.ToBase64String(myAsciiBytes);
			return myBase64String;
		}
		//=============================================================================================================

		/// TODO: Double Check the Accuracy of this, and the need for the CRLF translation
		/// <summary>Decode BASE64 String to ASCII String</summary>
		/// <param name="theBase64String">The BASE64 String</param>
		/// <returns>Decoded ASCII String</returns>
		/// <remarks>
		/// The BASE64 Data may be in UNIX format with ASCII LF characters only instead of CR/LF.
		/// When an ASCII LF character is encountered, Insert an ASCII CR before it for MS DOS/Windows Use.
		/// </remarks>
		/// http://en.wikipedia.org/wiki/Base64
		public static String Base64_Decode(
			String theBase64String
			)
		{
			int iByte;							// Index to myAsciiBytes
			int nLF;							// Count of <LF> Characters
			int iLF;							// Copy Offset Adjustment due to <LF> Characters
			int nCR;							// Count of <CR> Characters
			int iCR;							// Copy Offset Adjustment due to <CR> Characters
			int nNUL;							// Count of <NUL> Characters
			int iNUL;							// Copy Offset Adjustment due to <NUL> Characters
			int nAscii;							// Count of ASCII Characters
			byte[] myAsciiBytes;				// RAW Translation from Base64 to ASCII
			int myAsciiBytesLength;				// Length of myAsciiBytes
			byte[] myAsciiBytesWithCRLF;		// Processed Translation to replace <CR> with NIL, <LF> with <CR><LF> for UNIX / DOS-WINDOWS Compatibility
			String myAsciiString;				// the Returned Value, normally a copy of myAsciiBytesWithCRLF

			// Convert Base64 to ASCII
			try
			{
				myAsciiBytes = Convert.FromBase64String(theBase64String);

				// Count Number of LineFeeds, CarriageReturns, ZERO bytes, and ordinary ASCII Characters in myAsciiBytes[]
				myAsciiBytesLength = myAsciiBytes.GetLength(0);
				for (iByte = 0, nLF = 0, nCR = 0, nNUL = 0, nAscii = 0; iByte < myAsciiBytesLength; iByte++)
				{
					if (myAsciiBytes[iByte] == 0)
						nNUL++;
					else if (myAsciiBytes[iByte] == 10)
						nLF++;
					else if (myAsciiBytes[iByte] == 13)
						nCR++;
					else
						nAscii++;
				}

				// Allocate Another Buffer with Sufficient Additional Size to add a CR after each LF
				myAsciiBytesWithCRLF = new byte[nAscii + (2 * nLF)];

				// Copy Bytes from One Buffer to the Other, Adding CR before each LF, Skipping Each CR
				for (iByte = 0, iLF = 0, iCR = 0, iNUL = 0; iByte < myAsciiBytesLength; iByte++)
				{
					// Is it LF?
					if (myAsciiBytes[iByte] == 10)
					{
						// Add CR/LF
						myAsciiBytesWithCRLF[iByte + iLF - iCR - iNUL] = 13;
						iLF++;
						myAsciiBytesWithCRLF[iByte + iLF - iCR - iNUL] = 10;
					}

					// Is it CR?
					else if (myAsciiBytes[iByte] == 13)
					{
						// Skip CR
						iCR++;
					}

					// Is it ZERO?
					else if (myAsciiBytes[iByte] == 0)
					{
						// Skip Zero
						iNUL++;
					}

					// Is it Something Else?
					else
					{
						// Copy Byte
						myAsciiBytesWithCRLF[iByte + iLF - iCR - iNUL] = myAsciiBytes[iByte];
					}
				}

				// Convert byte[] to String
				myAsciiString = Encoding.ASCII.GetString(myAsciiBytesWithCRLF);
			}
			catch //(Exception e)
			{
				myAsciiString = theBase64String;
			}

			// Return String Value
			return myAsciiString;
		}
		//=============================================================================================================

		/// <summary>Search the XML REsponse String for the Return Value or Fault Response</summary>
		/// <param name="theXmlResponse">the XML String</param>
		/// <returns>Inner Text of First Node Found, or Empty String if Fault Response</returns>
		/// http://en.wikipedia.org/wiki/Base64
		/// http://en.wikipedia.org/wiki/XML-RPC 
		public String XmlRpcGetResponseValue(
			XmlDocument theXmlResponse
			)
		{

			XmlNode myParamsNode;				// Params Node in a Normal Response
			XmlNodeList myMemberNodeList;		// Member Node List in a Fault Response
			XmlNode myNameNode;					// Member Name Node in a Fault Response
			XmlNode myValueNode;				// Value Node in a Fault Response

			String myFaultCode = "";			// Fault Code (Fault Response)
			String myFaultString = "";			// Fault String (Fault Response)
			String myName = "";
			String myValue = "";

			try
			{
				// Check for Params Node (Normal Response)
				if (null != (myParamsNode = theXmlResponse.SelectSingleNode("/methodResponse/params")))
				{
					// Get the Inner Text of the XML Node
					//	Note: This simple approach works because there is ONLY ONE text value inside the XML
					myValue = myParamsNode.InnerText;
				}
				
				// Check for Fault Node (Error Response)
				else if (null != (myMemberNodeList = theXmlResponse.SelectNodes("/methodResponse/fault/value/struct/member")))
				{
					// Check Each Member in Structure for Information
					foreach (XmlElement myMemberNode in myMemberNodeList)
					{
						// Need to Use Relative Path Starting with "."
						myNameNode = myMemberNode.SelectSingleNode("./name");
						myName = myNameNode.InnerText;

						// Capture Fault Code
						if (myName == "faultCode")
						{
							// Note, the type is omitted from the XPATH since the inner text will be the same regardless
							myValueNode = myMemberNode.SelectSingleNode("./value");
							myFaultCode = myValueNode.InnerText;
						}

						// Capture Fault String
						if (myName == "faultString")
						{
							// Note, the type is omitted from the XPATH since the inner text will be the same regardless
							myValueNode = myMemberNode.SelectSingleNode("./value");
							myFaultString = myValueNode.InnerText;
						}

					}

					// Copy Fault Code and String to Error Properties
					m_ErrorCategory = FLDIGI_ErrorCategory.XML_RPC_ServerError;
					m_ErrorName = String.Format("XML-RPC-{0}", myFaultCode);
					m_ErrorMessage = myFaultString;

					// Force Value to Empty String
					myValue = String.Empty;
				}

				// This Should Never Happen, but Handle It Anyway
				else
				{
					// No Member Node List
					m_ErrorNumber = 111;
					m_ErrorCategory = FLDIGI_ErrorCategory.XML_RPC_ServerError;
					m_ErrorName = String.Format("XML-RPC{0}", 8000);
					m_ErrorMessage = "No Value or Error Information Present";

					// Force Value to Empty String
					myValue = String.Empty;
				}
			}
			catch (Exception e)
			{
				myValue = e.Message;
			}

			// Return the Value
			return myValue;
		}
		//=============================================================================================================

		/// <summary>Issue a Post to the HTTP Server</summary>
		/// <param name="thePost">The Post Data</param>
		/// <param name="theURL">URL of the HTTP Server</param>
		/// <returns>The Response Data</returns>
		/// Code Copied and Refactored from: http://msdn.microsoft.com/en-us/library/debx8sh9.aspx
		/// http://msdn.microsoft.com/en-us/library/System.Net.WebRequest(v=vs.100).aspx 
		/// http://msdn.microsoft.com/en-us/library/System.Net.WebResponse(v=vs.100).aspx
		/// http://msdn.microsoft.com/en-us/library/System.IO.StreamReader(v=vs.100).aspx
		/// http://msdn.microsoft.com/en-us/library/System.IO.Stream(v=vs.100).aspx
		/// http://msdn.microsoft.com/en-us/library/System.Text.Encoding(v=vs.100).aspx
		/// 
		public String HTTP_PostToServer(
			String thePost,
			String theURL
			)
		{
			String myResponseString;			// Returned Response String
			WebRequest myRequest;				// WEB Request
			WebResponse myResponse;				// WEB Response
			Stream myDataStream;				// I/O Data Stream
			StreamReader myStreamReader;		// Stream Reader
			byte[] mySendBuffer;				// Send Buffer

			// Post a Request to Server and Receive Response
			try
			{
				// Create a Request to the HTTP Server
				myRequest = WebRequest.Create(theURL);

				// Set the Method property of the request to POST.
				myRequest.Method = "POST";

				// Convert Post String to Byte Array
				mySendBuffer = Encoding.UTF8.GetBytes(thePost);

				// Set the ContentType property of the WebRequest.
				myRequest.ContentType = "application/x-www-form-urlencoded";

				// Set the ContentLength property of the WebRequest.
				myRequest.ContentLength = mySendBuffer.Length;

				// Get the Request Stream
				myDataStream = myRequest.GetRequestStream();

				// Write to Request Stream
				myDataStream.Write(mySendBuffer, 0, mySendBuffer.Length);

				// Close the Stream
				myDataStream.Close();
				//-------------------------------------------------------------------------------------

				// Get the Response
				myResponse = myRequest.GetResponse();

				// Get the Stream Containing the Response
				myDataStream = myResponse.GetResponseStream();

				// Open the Stream using a Stream Reader
				myStreamReader = new StreamReader(myDataStream);

				// Read the Entire Content
				myResponseString = myStreamReader.ReadToEnd();

				// Clean Up, Reader, Stream, WebResponse
				myStreamReader.Close();
				myDataStream.Close();
				myResponse.Close();
			}
			catch (Exception ex)
			{
				// Set Error Attributes
				m_ErrorNumber = 666;
				m_ErrorCategory = FLDIGI_ErrorCategory.NetworkError;
				m_ErrorName = "No Server Response";
				m_ErrorMessage = ex.Message;
				myResponseString = String.Empty;
			}

			// Return the Response String
			return myResponseString;
		}
		//=============================================================================================================

		/// <summary>Strip the XML Declaration from an XML Document String</summary>
		/// <param name="theXmlString">the XML Document String</param>
		/// <returns>the Stripped String</returns>
		private static String StripXmlDeclaration(
			String theXmlString
			)
		{
			int myIndex;
			String myString;

			// Find the ">" at the end of the XML Declaration
			myIndex = theXmlString.IndexOf(">");

			// Keep only the Substring Following that
			myString = theXmlString.Substring(myIndex + 1);

			return myString;
		}
		//=============================================================================================================

		/// <summary>Issue Request to FLDIGI XML-RPC Server and Get XML Response</summary>
		/// <param name="theXmlQuery">the XML Query</param>
		/// <returns>the XML Response from the XML-RPC Server</returns>
		/// http://en.wikipedia.org/wiki/XML-RPC
		public XmlDocument XmlRpcQueryServer(
			XmlDocument theXmlQuery
			)
		{
			XmlDocument myXmlResponse;
			String myResponse;
			int myFaultCode;
			String myFaultMessage;
			String myXmlQueryString;

			myXmlResponse = new XmlDocument();
			try
			{
				// Debugging the Response?
				if (m_SimulatedResponseFilename == String.Empty)
				{
					// Debugging the Query?
					if (m_SimulatedQueryFilename != String.Empty)
					{
						// Yes, Debugging, Use Simulated Query File
						// Check if File Exists
						if (File.Exists(m_SimulatedQueryFilename))
						{
							// Yes, Load File into String
							// Load Directly to String Used to allow Simulation of XML Errors in Simulated Queries
							myXmlQueryString = File.ReadAllText(m_SimulatedQueryFilename);
						}
						else
						{
							// No, Use Empty String
							myXmlQueryString = "<QUERY>Non-Existent File</QUERY>";
							m_ErrorNumber = 2;
							m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
							m_ErrorName = "Simulated Query File Doesn't Exist";
							m_ErrorMessage = Path.GetFileName(m_SimulatedQueryFilename);
						}
					}
					else
					{
						// No, Post Request to FLDIGI XML-RPC Server and Return the Response
						myXmlQueryString = theXmlQuery.OuterXml;
					}

					// Log Query
					if (m_QueryResponseLogFilename != String.Empty)
					{
						File.AppendAllText(m_QueryResponseLogFilename, String.Format("\r\n<QUERY>{0}</QUERY>\r\n", DateTime.Now));
						File.AppendAllText(m_QueryResponseLogFilename, FLDIGI.StripXmlDeclaration(myXmlQueryString));
					}

					// Is Query String Empty?
					if (myXmlQueryString == String.Empty)
					{
						// No, Use Empty String
						myXmlQueryString = String.Empty;
						m_ErrorNumber = 3;
						m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
						m_ErrorName = "Query String is Empty";
						m_ErrorMessage = String.Empty;
						myResponse = String.Empty;
					}
					else
					{
						// Post Request to FLDIGI XML-RPC Server and Return the Response
						myResponse = HTTP_PostToServer(myXmlQueryString, m_XmlRpcServerName);
					}

					// Log Response
					if (m_QueryResponseLogFilename != String.Empty)
					{
						File.AppendAllText(m_QueryResponseLogFilename, String.Format("\r\n<RESPONSE>{0}</RESPONSE>\r\n", DateTime.Now));
						File.AppendAllText(m_QueryResponseLogFilename, FLDIGI.StripXmlDeclaration(myResponse));
					}

					// Load Response into XML Document
					if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
					{
						// Normal Response
						myXmlResponse.LoadXml(myResponse);
					}
					else
					{
						// Error Response in XML-RPC Format
						myFaultCode = m_ErrorNumber;
						myFaultMessage = String.Format("{0}: {1} - {2}", m_ErrorName, m_ErrorCategory, m_ErrorMessage);
						XmlRpcAppendDeclaration(myXmlResponse);
						XmlRpcAppendFaultResponse(myXmlResponse, myFaultCode, myFaultMessage);
					}
				}
				else
				{
					// Debugging, use Simulated Response File
					myXmlResponse.Load(m_SimulatedResponseFilename);

					if (File.Exists(m_SimulatedResponseFilename))
					{
						// Yes, Load File into String
						// Load Directly to String Used to allow Simulation of XML Errors in Simulated Queries
						myXmlResponse.Load(m_SimulatedResponseFilename);
					}
					else
					{
						// No, Use Empty String
						myXmlQueryString = String.Empty;
						m_ErrorNumber = 3;
						m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
						m_ErrorName = "Simulated Response File Doesn't Exist";
						m_ErrorMessage = Path.GetFileName(m_SimulatedQueryFilename);
					}
				}

			}
			catch (Exception ex)
			{
				m_ErrorNumber = 777;
				m_ErrorCategory = FLDIGI_ErrorCategory.XML_Error;
				m_ErrorName = "XML";
				m_ErrorMessage = ex.Message;
			}

			return myXmlResponse;
		}
		//=========================================================================================

		/// <summary>Append XML Declaration to Empty XML Document</summary>
		/// <param name="theXmlDocument">The XML Document, New but Empty</param>
		/// <returns>XML String Containing the Declaration, indended for Debugging</returns>
		/// <remarks>
		/// theXmlDocument must be New but Empty, and is modified to contain the XML declaration.
		/// </remarks>
		/// http://msdn.microsoft.com/en-us/library/system.xml.xmldocument.createxmldeclaration(v=vs.100).aspx
		public String XmlRpcAppendDeclaration(
			XmlDocument theXmlDocument
			)
		{
			// Create XML Declaration
			String myXmlVersion = "1.0";			// Default
			String myXmlEncoding = "utf-8";			// Default, Original Example
			//String myXmlEncoding = "iso-8859-1";	// Used by QRZ.com
			String myXmlStandAlone = "";			// ???
			XmlDeclaration myXmlDeclaration;
			String myDeclarationString;

			// Add Declaration to XML Document
			myXmlDeclaration = theXmlDocument.CreateXmlDeclaration(myXmlVersion, myXmlEncoding, myXmlStandAlone);
			theXmlDocument.PrependChild(myXmlDeclaration);
			myDeclarationString = myXmlDeclaration.OuterXml;

			return myDeclarationString;
		}
		//=========================================================================================

		/// <summary>Append Method Call to XML Document</summary>
		/// <param name="theXmlDocument">The XML Document Containing Only a Declaration</param>
		/// <param name="theMethod">The Method Name</param>
		/// <returns>XML String Containing the Method Call, intended for Debugging</returns>
		/// <remarks>
		/// theXmlDocument Must Contain Only a Declaration, and is Modified to contain the XML Method Call.
		/// </remarks>
		/// http://msdn.microsoft.com/en-us/library/System.Xml.XmlDocument.CreateElement(v=vs.100).aspx
		/// http://en.wikipedia.org/wiki/XML-RPC
		public String XmlRpcAppendMethodCall(
			XmlDocument theXmlDocument,
			String theMethod
			)
		{
			XmlNode myMethodCall;
			XmlNode myMethodName;
			XmlNode myParams;

			// Create MethodCall
			myMethodCall = theXmlDocument.CreateElement("methodCall");
			theXmlDocument.AppendChild(myMethodCall);

			// Create MethodName
			myMethodName = theXmlDocument.CreateElement("methodName");
			myMethodCall.AppendChild(myMethodName);
			myMethodName.InnerText = theMethod;

			// Create Params
			myParams = theXmlDocument.CreateElement("params");
			myMethodCall.AppendChild(myParams);

			String myMethodCallString = myMethodCall.OuterXml;

			// Test
			return myMethodCallString;
		}
		//=========================================================================================

		/// <summary>Append Parameter to XML Document</summary>
		/// <param name="theXmlDocument">The XML Document Containing a Method Call</param>
		/// <param name="theValue">the Argument to be Appended</param>
		/// <returns></returns>
		public String XmlRpcAppendParamValue(
			XmlDocument theXmlDocument,
			XmlNode theValue)
		{
			XmlNode myParams;
			XmlNode myParam;
			XmlNode myValue;
			String myParamsString;

			// Find the params Node
			myParams = theXmlDocument.SelectSingleNode("./methodCall/params");

			// Create Parameter
			myParam = theXmlDocument.CreateElement("param");
			myParams.AppendChild(myParam);

			// Create Value
			myValue = theXmlDocument.CreateElement("value");
			myParam.AppendChild(myValue);

			// Append Param Value
			myValue.AppendChild(theValue);

			myParamsString = myParams.OuterXml;

			return myParamsString;
		}
		//=========================================================================================

		/// <summary>Append Parameter to XML-RPC Method call</summary>
		/// <param name="theXmlDocument">An Existing XML Document Containing an XML-RPC Method Call</param>
		/// <param name="theValue">The Parameter Value</param>
		/// <param name="theType">The Parameter Type</param>
		/// <returns>String Containing XML for the Parameter, intended for Debugging</returns>
		/// <remarks>
		/// theXmlDocument must Contain a Declaration and Method Call, and is Modified to Contain the Parameter.
		/// </remarks>
		/// http://msdn.microsoft.com/en-us/library/System.Xml.XmlNode.SelectSingleNode(v=vs.100).aspx
		/// http://en.wikipedia.org/wiki/XML-RPC
		public String XmlRpcAppendParam(
			XmlDocument theXmlDocument,
			String theValue,
			String theType
			)
		{
			XmlNode myParams;
			XmlNode myParam;
			XmlNode myValue;
			XmlNode myType;
			String myParamsString;

			// Find the params Node
			myParams = theXmlDocument.SelectSingleNode("/methodCall/params");

			// Create Parameter
			myParam = theXmlDocument.CreateElement("param");
			myParams.AppendChild(myParam);

			// Create Value
			myValue = theXmlDocument.CreateElement("value");
			myParam.AppendChild(myValue);

			// Create Parameter Type
			myType = theXmlDocument.CreateElement(theType);
			myValue.AppendChild(myType);

			// Add Text Value
			myType.InnerText = theValue;

			myParamsString = myParams.OuterXml;

			return myParamsString;
		}
		//=========================================================================================

		/// <summary>Create an XML Fault Response</summary>
		/// <param name="theXmlDocument">An Existing XML Document with a Declaration</param>
		/// <param name="theFaultCode">The Numeric Fault Code</param>
		/// <param name="theFaultMessage">The Fault Message</param>
		/// <returns>XML Fault Response with Fault Code and Fault Message, intended for Debugging</returns>
		/// <remarks>This Function is Used to Synthesize an XML Fault Response When an Exception Occurs Due to Network, XML, or Other Reasons</remarks>
		/// http://en.wikipedia.org/wiki/XML-RPC
		public String XmlRpcAppendFaultResponse(
			XmlDocument theXmlDocument,
			int theFaultCode,
			String theFaultMessage
			)
		{
			String myFaultResponseString;
			XmlNode myMethodResponse;
			XmlNode myFault;
			XmlNode myValue;
			XmlNode myStruct;
			XmlNode myMember;
			XmlNode myName;
			XmlNode myInt;
			XmlNode myString;
	
			// Create and Append the MethodResponse Node
			myMethodResponse = theXmlDocument.CreateElement("methodResponse");
			theXmlDocument.AppendChild(myMethodResponse);

			// Create and Append the Fault Node
			myFault = theXmlDocument.CreateElement("fault");
			myMethodResponse.AppendChild(myFault);

			// Create and Append the Top Level Value Node
			myValue = theXmlDocument.CreateElement("value");
			myFault.AppendChild(myValue);

			// Create and Append the Struct Node
			myStruct = theXmlDocument.CreateElement("struct");
			myValue.AppendChild(myStruct);


			// Create and Append the Fault Code Member Node
			myMember = theXmlDocument.CreateElement("member");
			myStruct.AppendChild(myMember);

			// Create and Append the Fault Code Name Node
			myName = theXmlDocument.CreateElement("name");
			myMember.AppendChild(myName);
			myName.InnerText = "faultCode";

			// Create and Append the Fault Code Value Node
			myValue = theXmlDocument.CreateElement("value");
			myMember.AppendChild(myValue);

			// Create and Append the Fault Code Type Node
			myInt = theXmlDocument.CreateElement("int");
			myValue.AppendChild(myInt);
			myInt.InnerText = String.Format("{0}", theFaultCode);


			// Create and Append the Fault String Member Node
			myMember = theXmlDocument.CreateElement("member");
			myStruct.AppendChild(myMember);

			// Create and Append the Fault String Name Node
			myName = theXmlDocument.CreateElement("name");
			myMember.AppendChild(myName);
			myName.InnerText = "faultString";

			// Create and Append the Fault String Value Node
			myValue = theXmlDocument.CreateElement("value");
			myMember.AppendChild(myValue);

			// Create and Append the Fault String Type Node
			myString = theXmlDocument.CreateElement("string");
			myValue.AppendChild(myString);
			myString.InnerText = String.Format("{0}", theFaultMessage);

			myFaultResponseString = myMethodResponse.OuterXml;
			return myFaultResponseString;
		}
		//=========================================================================================

		///////////////////////////////////////////////////////////////////////////////////////////
		//	XML-RPC METHOD FAMILIES BY SIGNATURE
		//	Naming Convention is "XmlRpc" followed by "_" <return type> "_" <argument types>
		//	As described in the FLDIGI XML-RPC Documentation
		//	Example: Method "fldigi.name" has signature "s:n"
		//		<return type> is "s", <argument types> is "n"
		//		Signature Family Function is "String XmlRpc_s_n()"
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>XML-RPC Function: Nil Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		public void XmlRpc_n_n(
			String theFunction
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			 myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			// Void
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Boolean Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>String Value</returns>
		public Boolean XmlRpc_b_n(
			String theFunction
			)
		{
			Boolean myValue;					// Returned Boolean value
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = "0";
			}

			// Decode Value
			if (myValueString == "1")
				myValue = true;
			else if (myValueString == "0")
				myValue = false;
			else
			{
				myValue = false;

				// Indicate Error
				m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
				m_ErrorName = "DATA-B1";
				m_ErrorMessage = String.Format("Error Converting Return Value {0} to Boolean", myValueString);
			}


			// Return Value
			return myValue;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Integer Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>Integer Value</returns>
		public Int32 XmlRpc_i_n(
			String theFunction
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = "0";
			}


			// Decode Value String
			Int32 myValue;
			try
			{
				myValue = Convert.ToInt32(myValueString);
			}
			catch //(Exception e)
			{
				// Dummy Value
				myValue = 0;

				// Indicate Error
				m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
				m_ErrorName = "DATA-I32a";
				m_ErrorMessage = String.Format("Error Converting Return Value {0} to Int32", myValueString);
			}

			// Return Value
			return myValue;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Double Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>Double Value</returns>
		public Double XmlRpc_d_n(
			String theFunction
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = "0.0";
			}

			// Decode Value String
			Double myValue;
			try
			{
				myValue = Convert.ToDouble(myValueString);
			}
			catch //(Exception e)
			{
				myValue = 0.0;

				// Indicate Error
				m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
				m_ErrorName = "DATA-D64a";
				m_ErrorMessage = String.Format("Error Converting Return Value {0} to Double", myValueString);
			}

			// Return Value
			return myValue;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: String Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>String Value</returns>
		public String XmlRpc_s_n(
			String theFunction
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			return myValueString;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Structure Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>String Value</returns>
		public String XmlRpc_S_n(
			String theFunction
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				XmlNode myStructure = myXmlResponse.SelectSingleNode("/methodResponse/params/param/value");
				myValueString = myStructure.OuterXml;
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			return myValueString;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Nil Return Value, Array or Structure Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theXmlMethodCall">An Empty XML Document to Hold an XML Method Call</param>
		/// <param name="theArgument">An Array or Structure Value in XML-RPC Format</param>
		public void XmlRpc_n_V(
			String theFunction,
			XmlDocument theXmlMethodCall,
			XmlNode theArgument
			)
		{
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Add XML Declaration
			XmlRpcAppendDeclaration(theXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(theXmlMethodCall, theFunction);

			// Add Argument to XML Method Call
			XmlRpcAppendParamValue(theXmlMethodCall, theArgument);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(theXmlMethodCall);
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Array or Structure Return Value, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns></returns>
		public XmlDocument XmlRpc_V_n(
			String theFunction
			)
		{
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Return Entire Response
			return myXmlResponse;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: String Array Return, Nil Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>String Array</returns>
		public String[] XmlRpc_A_n(
			String theFunction
			)
		{
			String[] myStringArray;				// Returned Array
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response
			XmlNodeList myNodeList;				// Array NodeList

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Debug Probe
			myValueString = myXmlResponse.OuterXml;

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				// Get List of Values in Array
				myNodeList = myXmlResponse.SelectNodes("/methodResponse/params/param/value/array/data/value");
				int nNodes = myNodeList.Count;
				if (nNodes > 0)
				{
					int i = 0;
					myStringArray = new String[nNodes];
					foreach (XmlNode myXmlNode in myNodeList)
					{
						// Extract Value String from Each Value
						myStringArray[i++] = myXmlNode.InnerText;
					}
				}
				else
				{
					// Null Value if NO Array Elements
					myStringArray = null;
				}
			}
			else
			{
				myStringArray = null;
			}

			// Return Value
			return myStringArray;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Nil Return, Boolean Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">A Boolean Value Argument</param>
		public void XmlRpc_n_b(
			String theFunction,
			Boolean theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response
			String myArgument;					// Argument

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			myArgument = (theArgument == true ? "1" : "0");

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "boolean");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}


			// Return Value
			// Void
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Integer Return, Boolean Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">an Integer Argument</param>
		public void XmlRpc_n_i(
			String theFunction,
			Int32 theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			String myArgument = String.Format("{0}", theArgument);

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "int");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			// Void
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Nil Return, String Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">a String Argument</param>
		public void XmlRpc_n_s(
			String theFunction,
			String theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, theArgument, "string");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			// Void
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Nil Return, String Array Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">a String Array Argument</param>
		public void XmlRpc_n_A(
			String theFunction,
			String[] theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// TODO: Finish Implementation of Type A Argument
			// Append Parameter to XML Method Call

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			// Void
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: String Return, String Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">String Argument</param>
		/// <returns></returns>
		public String XmlRpc_s_s(
			String theFunction,
			String theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, theArgument, "string");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			return myValueString;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: String Return, Integer Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">Integer Argument</param>
		/// <returns>String Value</returns>
		public String XmlRpc_s_i(
			String theFunction,
			Int32 theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			String myArgument = String.Format("{0}", theArgument);

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "int");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			return myValueString;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: String Return, Boolean Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">Boolean Argument</param>
		/// <returns>String Value</returns>
		public String XmlRpc_s_b(
			String theFunction,
			Boolean theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			String myArgument = String.Format("{0}", theArgument);

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "boolean");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			return myValueString;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Integer Return, Integer Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">Integer Argument</param>
		/// <returns></returns>
		public Int32 XmlRpc_i_i(
			String theFunction,
			Int32 theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			String myArgument = String.Format("{0}", theArgument);

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "int");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = "0";
			}

			// Decode Value
			Int32 myValue;
			try
			{
				myValue = Convert.ToInt32(myValueString);
			}
			catch //(Exception e)
			{
				myValue = 0;

				// Indicate Error
				m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
				m_ErrorName = "DATA-I32b";
				m_ErrorMessage = String.Format("Error Converting Return Value {0} to Int32", myValueString);
			}

			// Return Value
			return myValue;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Boolean Return, Boolean Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">Boolean Argument</param>
		/// <returns></returns>
		public Boolean XmlRpc_b_b(
			String theFunction,
			Boolean theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response
			String myArgument;					// Argument

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			myArgument = (theArgument == true ? "1" : "0");

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "boolean");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = "0";
			}

			// Decode Value
			Boolean myValue;
			if (myValueString == "1")
				myValue = true;
			else if (myValueString == "0")
				myValue = false;
			else
			{
				myValue = false;

				// Indicate Error
				m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
				m_ErrorName = "DATA-B2";
				m_ErrorMessage = String.Format("Error Converting Return Value {0} to Boolean", myValueString);
			}

			// Return Value
			return myValue;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Double Return, Double Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">Double Argument</param>
		/// <returns>Double Value</returns>
		public Double XmlRpc_d_d(
			String theFunction,
			Double theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response
			String myArgument;					// Argument

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			myArgument = String.Format("{0}", theArgument);

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "double");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = "0.0";
			}

			// Decode Value
			Double myValue;
			try
			{
				myValue = Convert.ToDouble(myValueString);
			}
			catch //(Exception e)
			{
				myValue = 0.0;

				// Indicate Error
				m_ErrorCategory = FLDIGI_ErrorCategory.DataError;
				m_ErrorName = "DATA-D64b";
				m_ErrorMessage = String.Format("Error Converting Return Value {0} to Double", myValueString);
			}

			// Return Value
			return myValue;
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Nil Return, Base64 Argument</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument">String Argument</param>
		public void XmlRpc_n_6(
			String theFunction,
			String theArgument
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response
			String myArgument;					// Argument

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Translate Argument to String
			myArgument = FLDIGI.Base64_Encode(theArgument);

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument, "base64");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return Value
			// Void
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Base64 Return, Two Integer Arguments</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <param name="theArgument1">Integer Argument</param>
		/// <param name="theArgument2">Integer Argument</param>
		/// <returns>String Value in BASE64 Format</returns>
		public String XmlRpc_6_ii(
			String theFunction,
			Int32 theArgument1,
			Int32 theArgument2
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Translate Parameters to String
			String myArgument1 = String.Format("{0}", theArgument1);
			String myArgument2 = String.Format("{0}", theArgument2);

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument1, "int");

			// Append Parameter to XML Method Call
			XmlRpcAppendParam(myXmlMethodCall, myArgument2, "int");

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return valule
			return Base64_Decode(myValueString);
		}
		//=========================================================================================

		/// <summary>XML-RPC Function: Base64 Return, Nil Arguments</summary>
		/// <param name="theFunction">The Function Name from FLDIGI XML-RPC Documentation</param>
		/// <returns>String Value in BASE64 Format</returns>
		public String XmlRpc_6_n(
			String theFunction
			)
		{
			String myValueString;				// Returned Value String
			XmlDocument myXmlMethodCall;		// XML Method Call
			XmlDocument myXmlResponse;			// XML Response

			// Reset Error Code
			m_ErrorCategory = FLDIGI_ErrorCategory.OK;

			// Create New XML Method Call
			myXmlMethodCall = new XmlDocument();

			// Add XML Declaration
			XmlRpcAppendDeclaration(myXmlMethodCall);

			// Populate XML Method Call
			XmlRpcAppendMethodCall(myXmlMethodCall, theFunction);

			// Issue Request
			myXmlResponse = XmlRpcQueryServer(myXmlMethodCall);

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				myValueString = XmlRpcGetResponseValue(myXmlResponse);
			}
			else
			{
				myValueString = String.Empty;
			}

			// Return valule
			return Base64_Decode(myValueString);
		}
		//=========================================================================================

		////////////////////////////////////////////////////////////////////////////////////////////
		//	XML-RPC METHOD FUNCTIONS
		//	Naming Convention is as follows:
		//	Method Name in FLDIGI Documentation is similar to Function Name
		//		1) Each Word Has First Letter Capitalized
		//		2) Dots/ Period is converted to Underscore "_"
		//		3) Underscore remains Underscore "_"
		//		4) Acronyms such as ID or AFC have ALL Letters Capitalized
		//	Example: "fldigi.name_version" is "String Fldigi_Name_Version()"
		///////////////////////////////////////////////////////////////////////////////////////////

		// FLDIGI FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>fldigi.config_dir - s:n - Returns the FLDIGI Configuration Directory</summary>
		/// <returns>FLDIGI Configuration Directory</returns>
		public String Fldigi_Config_Dir()
		{
			return XmlRpc_s_n("fldigi.config_dir");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.list - A:n - Returns the FLDIGI List of Methods</summary>
		/// <returns>FLDIGI List of Methods</returns>
		/// <remarks>
		/// Note, the Return is a String Array
		/// Each String is the XmlDocument.InnerText which is from a structure
		/// </remarks>
		public String[] Fldigi_List()
		{
			//return XmlRpc_A_n("fldigi.list");

			XmlDocument myXmlResponse;			// Complete XML-RPC Response
			XmlNodeList myXmlNodeList;			// Data Value Node List
			XmlNodeList myMemberList;			// Structure Member List
			XmlNode myNameNode;					// Member Name Node
			XmlNode myValueNode;				// Member Value Node
			String[] myStringArray;				// Returned Array
			String myMethodName = "";
			String myMethodSignature = "";
			String myHelp = "";

			myXmlResponse = XmlRpc_V_n("fldigi.list");

			// TODO: Decode Structure for each Array Element to Make Better List
			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				// Get List of Values in Array
				myXmlNodeList = myXmlResponse.SelectNodes("/methodResponse/params/param/value/array/data/value");
				int nNodes = myXmlNodeList.Count;
				if (nNodes > 0)
				{
					int i = 0;
					myStringArray = new String[nNodes];
					foreach (XmlNode myXmlNode in myXmlNodeList)
					{
						// Extract Value String from Each Value
						//myStringArray[i++] = myXmlNode.InnerText;

						// Get List of Structure Members
						myMemberList = myXmlNode.SelectNodes("./struct/member");
						foreach (XmlNode myMember in myMemberList)
						{
							myNameNode = myMember.SelectSingleNode("./name");
							myValueNode = myMember.SelectSingleNode("./value");
							switch (myNameNode.InnerText)
							{
								case "help":
									myHelp = myValueNode.InnerText;
									break;
								case "name":
									myMethodName = myValueNode.InnerText;
									break;
								case "signature":
									myMethodSignature = myValueNode.InnerText;
									break;
								default:
									break;
							}
						}

						// Format Items in Structure for ComboBox
						myStringArray[i++] = String.Format("{0} - {1} - {2}", myMethodName, myMethodSignature, myHelp);
					}
				}
				else
				{
					// Null Value if NO Array Elements
					myStringArray = null;
				}
			}
			else
			{
				myStringArray = null;
			}

			// Return Value
			return myStringArray;
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.name - s:n - Returns the FLDIGI Program Name</summary>
		/// <returns>FLDIGI Program Name</returns>
		public String Fldigi_Name()
		{
			return XmlRpc_s_n("fldigi.name");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.name_version - s:n - Returns the FLDIGI Program Name and Version</summary>
		/// <returns>FLDIGI Program Name and Version</returns>
		public String Fldigi_Name_Version()
		{
			return XmlRpc_s_n("fldigi.name_version");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.terminate - n:i - Terminates FLDIGI Without Saving Anything</summary>
		public void Fldigi_Terminate()
		{
			XmlRpc_n_i("fldigi.terminate", 0);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.terminate - n:i - Terminates FLDIGI with Option Code</summary>
		/// <param name="theTerminationOption">Termination Option Code</param>
		/// <remarks>
		/// Termination Option Code is bitmask specifying data to save: 
		/// bit 0=options; 
		/// bit 1=log; 
		/// bit 2=macros.
		/// </remarks>
		public void Fldigi_Terminate(
			int theTerminationOption
			)
		{
			XmlRpc_n_i("fldigi.terminate", theTerminationOption);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.version - s:n - Returns the FLDIGI Program Version</summary>
		/// <returns>FLDIGI Program Version</returns>
		public String Fldigi_Version()
		{
			return XmlRpc_s_n("fldigi.version");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>fldigi.version_struct - S:n - Returns the FLDIGI Program Version Structure</summary>
		/// <returns>FLDIGI Program Version Structure</returns>
		/// <remarks>Returned Value is Outer XML of Value</remarks>
		public String Fldigi_Version_Struct()
		{
			XmlDocument myXmlResponse;
			String myValueString;

			//return XmlRpc_S_n("fldigi.version_struct");
			myXmlResponse = XmlRpc_V_n("fldigi.version_struct");

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				XmlNode myStructure = myXmlResponse.SelectSingleNode("/methodResponse/params/param/value");
				//myValueString = myStructure.OuterXml;
				myValueString = myStructure.OuterXml;
			}
			else
			{
				myValueString = String.Empty;
			}

			return myValueString;
		}
		//-----------------------------------------------------------------------------------------

		// LOG FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>log.clear - n:n - LOG / Clears ALL LOG Fields</summary>
		public void Log_Clear()
		{
			XmlRpc_n_n("log.clear");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_az - s:n - LOG / Get AZ</summary>
		/// <returns>AZ</returns>
		public String Log_Get_AZ()
		{
			return XmlRpc_s_n("log.get_az");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_band - s:n - LOG / Get Band Name</summary>
		/// <returns>Band Name</returns>
		public String Log_Get_Band()
		{
			return XmlRpc_s_n("log.get_band");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_call - s:n - LOG / Get Call</summary>
		/// <returns>Call</returns>
		public String Log_Get_Call()
		{
			return XmlRpc_s_n("log.get_call");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_country - s:n - LOG / Get Country String</summary>
		/// <returns>Country String</returns>
		public String Log_Get_Country()
		{
			return XmlRpc_s_n("log.get_country");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_exchange - s:n - LOG / Get Contest Exchange</summary>
		/// <returns>Contest Exchange</returns>
		public String Log_Get_Exchange()
		{
			return XmlRpc_s_n("log.get_exchange");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_frequency - s:n - LOG / Get Frequency</summary>
		/// <returns>LOG / Frequency</returns>
		public String Log_Get_Frequency()
		{
			return XmlRpc_s_n("log.get_frequency");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_locator - s:n - LOG / Get Locator (Maidenhead Grid)</summary>
		/// <returns>Locator</returns>
		public String Log_Get_Locator()
		{
			return XmlRpc_s_n("log.get_locator");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_name - s:n - LOG / Get Name</summary>
		/// <returns>Name</returns>
		public String Log_Get_Name()
		{
			return XmlRpc_s_n("log.get_name");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_notes - s:n - LOG / Get Notes</summary>
		/// <returns>Notes</returns>
		public String Log_Get_Notes()
		{
			return XmlRpc_s_n("log.get_notes");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_province - s:n - LOG / Get Province</summary>
		/// <returns>Province</returns>
		public String Log_Get_Province()
		{
			return XmlRpc_s_n("log.get_province");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_QTH - s:n - LOG / Get QTH</summary>
		/// <returns>QTH</returns>
		public String Log_Get_QTH()
		{
			return XmlRpc_s_n("log.get_qth");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_rst_in - s:n - LOG / Get RST IN (Received)</summary>
		/// <returns>RST IN</returns>
		public String Log_Get_RST_In()
		{
			return XmlRpc_s_n("log.get_rst_in");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_rst_out - s:n - LOG / Get RST OUT (Sent)</summary>
		/// <returns>RST OUT</returns>
		public String Log_Get_RST_Out()
		{
			return XmlRpc_s_n("log.get_rst_out");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_serial_number - s:n - LOG / Get Serial Number</summary>
		/// <returns>Serial Number</returns>
		public String Log_Get_Serial_Number()
		{
			return XmlRpc_s_n("log.get_serial_number");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_serial_number_sent - s:n - LOG / Get Serial Number Sent</summary>
		/// <returns>Serial Number Sent</returns>
		public String Log_Get_Serial_Number_Sent()
		{
			return XmlRpc_s_n("log.get_serial_number_sent");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_state - s:n - LOG / Get State</summary>
		/// <returns>State</returns>
		public String Log_Get_State()
		{
			return XmlRpc_s_n("log.get_state");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_time_off - s:n - LOG / Get Time OFF</summary>
		/// <returns>Time OFF</returns>
		public String Log_Get_Time_Off()
		{
			return XmlRpc_s_n("log.get_time_off");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.get_time_on - s:n - LOG / Get Time ON</summary>
		/// <returns>Time OFF</returns>
		public String Log_Get_Time_On()
		{
			return XmlRpc_s_n("log.get_time_on");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.set_call - n:s - LOG / Set Call</summary>
		/// <param name="theCall">Call</param>
		public void Log_Set_Call(
			String theCall
			)
		{
			XmlRpc_n_s("log.set_call", theCall);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.set_exchange - n:s - LOG / Set Exchange</summary>
		/// <param name="theExchange">Exchange</param>
		public void Log_Set_Exchange(
			String theExchange
			)
		{
			XmlRpc_n_s("log.set_exchange", theExchange);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.set_locator - n:s - LOG / Set Locator (Maidenhead Grid)</summary>
		/// <param name="theLocator">Locator</param>
		public void Log_Set_Locator(
			String theLocator
			)
		{
			XmlRpc_n_s("log.set_locator", theLocator);
		}

		/// <summary>log.set_name - n:s - LOG / Set Name</summary>
		/// <param name="theName">Name</param>
		public void Log_Set_Name(
			String theName
			)
		{
			XmlRpc_n_s("log.set_name", theName);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.set_qth - n:s - LOG / Set QTH</summary>
		/// <param name="theQTH">QTH</param>
		public void Log_Set_QTH(
			String theQTH
			)
		{
			XmlRpc_n_s("log.set_qth", theQTH);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>log.set_serial_number - n:s - LOG / Set Serial Number</summary>
		/// <param name="theSerialNumber">Serial Number</param>
		public void Log_Set_Serial_Number(
			String theSerialNumber
			)
		{
			XmlRpc_n_s("log.set_serial_number", theSerialNumber);
		}
		//-----------------------------------------------------------------------------------------

		// MAIN FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>main.abort - n:n - MAIN / Abort</summary>
		/// <remarks>Aborts Transmit or Tune and Enters Receive Mode</remarks>
		public void Main_Abort()
		{
			XmlRpc_n_n("main.abort");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_afc - b:n - MAIN / Get Main AFC Status</summary>
		/// <returns>AFC Status</returns>
		public Boolean Main_Get_AFC()
		{
			return XmlRpc_b_n("main.get_afc");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_frequency - d:n - MAIN /  Get Frequency</summary>
		/// <returns>Frequency</returns>
		public Double Main_Get_Frequency()
		{
			return XmlRpc_d_n("main.get_frequency");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_lock - b:n - MAIN / Get Lock Status</summary>
		/// <returns>Lock Status</returns>
		public Boolean Main_Get_Lock()
		{
			return XmlRpc_b_n("main.get_lock");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_max_macro_id - i:n - MAIN / Get Max Macro ID</summary>
		/// <returns>Max Macro ID</returns>
		public Int32 Main_Get_Max_Macro_ID()
		{
			return XmlRpc_i_n("main.get_max_macro_id");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_reverse - b:n - MAIN / Get Reverse Status</summary>
		/// <returns>Reverse status</returns>
		public Boolean Main_Get_Reverse()
		{
			return XmlRpc_b_n("main.get_reverse");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_rsid - b:n - MAIN /  Get RxID Status</summary>
		/// <returns>RxID Status</returns>
		public Boolean Main_Get_RSID()
		{
			return XmlRpc_b_n("main.get_rsid");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_squelch - b:n - MAIN / Get Squelch Status</summary>
		/// <returns>Squelch Status</returns>
		public Boolean Main_Get_Squelch()
		{
			return XmlRpc_b_n("main.get_squelch");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_squelch_level - d:n - MAIN / Get Squelch Level</summary>
		/// <returns>Squelch Level</returns>
		public Double Main_Get_Squelch_Level()
		{
			return XmlRpc_d_n("main.get_squelch_level");
		}

		/// <summary>main.get_status1 - s:n - MAIN / Get Status 1</summary>
		/// <returns>Status 1</returns>
		/// <remarks>Status 1 is Modem Name (Digital Mode) Dependent</remarks>
		public String Main_Get_Status1()
		{
			return XmlRpc_s_n("main.get_status1");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_status2 - s:n - MAIN / Get Main Status 2</summary>
		/// <returns>Status 2 </returns>
		/// <remarks>Status 2 is Modem Name (Digital Mode) Dependent</remarks>
		public String Main_Get_Status2()
		{
			return XmlRpc_s_n("main.get_status2");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_trx_state - s:n - MAIN / Get TRX State</summary>
		/// <returns>TRX State</returns>
		public String Main_Get_TRX_State()
		{
			return XmlRpc_s_n("main.get_trx_state");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_trx_status - s:n - MAIN / Get TRX Status</summary>
		/// <returns>TRX Status</returns>
		public String Main_Get_TRX_Status()
		{
			return XmlRpc_s_n("main.get_trx_status");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.get_wf_sideband - s:n - MAIN / Get WF Sideband</summary>
		/// <returns>WF Sideband</returns>
		public String Main_Get_WF_Sideband()
		{
			return XmlRpc_s_n("main.get_wf_sideband");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.inc_frequency - d:d - MAIN / Increment RF Carrier Frequency, Returns New RF Carrier Frequency</summary>
		/// <param name="theFrequency">RF Carrier Frequency Increment</param>
		/// <returns>New RF Carrier Frequency</returns>
		public Double Main_Inc_Frequency(
			double theFrequency
			)
		{
			return XmlRpc_d_d("main.inc_frequency", theFrequency);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.inc_squelch_level - d:d - MAIN / Increments the Squelch Level, Returns New Squelch Level</summary>
		/// <param name="theSquelchLevelIncrement">Squelch Level Increment</param>
		/// <returns>New Squelch Level</returns>
		public Double Main_Inc_Squelch_Level(
			Double theSquelchLevelIncrement
			)
		{
			return XmlRpc_d_d("main.inc_squelch_level", theSquelchLevelIncrement);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.run_macro - n:i - MAIN / Run Macro</summary>
		/// <param name="theMacroID">Macro ID Number</param>
		public void Main_Run_Macro(
			Int32 theMacroID
			)
		{
			XmlRpc_n_i("main.run_macro", theMacroID);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.rx - n:n - MAIN / Enter Receive Mode</summary>
		public void Main_RX()
		{
			XmlRpc_n_n("main.rx");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_afc - b:b - MAIN / Set AFC status</summary>
		/// <param name="theAFC_Status">AFC status</param>
		/// <returns>Old AFC Status</returns>
		public Boolean Main_Set_AFC(
			Boolean theAFC_Status
			)
		{
			return XmlRpc_b_b("main.set_afc", theAFC_Status);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_frequency - d:d - MAIN / Set Frequency</summary>
		/// <param name="theFrequency">Frequency</param>
		/// <returns>Old Frequency</returns>
		public Double Main_Set_Frequency(
			double theFrequency
			)
		{
			return XmlRpc_d_d("main.set_frequency", theFrequency);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_lock - b:b - MAIN / Set Transmit Lock Status, Return Old Transmit Lock Status</summary>
		/// <param name="TheLockStatus">Transmit Lock Status</param>
		/// <returns>Old Transmit Lock Status</returns>
		public Boolean Main_Set_Lock(
			Boolean TheLockStatus
			)
		{
			return XmlRpc_b_b("main.set_lock", TheLockStatus);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_reverse - b:b -  MAIN Set Reverse Sideband Status, Return Old Reverse Sideband Status</summary>
		/// <param name="theReverseStatus">Reverse Sideband Status</param>
		/// <returns>Old Reverse Sideband Status</returns>
		public Boolean Main_Set_Reverse(
			Boolean theReverseStatus
			)
		{
			return XmlRpc_b_b("main.set_reverse", theReverseStatus);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_rsid - b:b - MAIN / Set RxID Status, Return Old RxID Status</summary>
		/// <param name="theRSID_Status">RxID Status</param>
		/// <returns>Old RxID Status</returns>
		public Boolean Main_Set_RSID(
			Boolean theRSID_Status
			)
		{
			return XmlRpc_b_b("main.set_rsid", theRSID_Status);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_squelch - b:b - MAIN / Set Squelch Status, Return Old Squelch Status</summary>
		/// <param name="theSquelchStatus">Squelch Status</param>
		/// <returns>Old Squelch Status</returns>
		public Boolean Main_Set_Squelch(
			Boolean theSquelchStatus
			)
		{
			return XmlRpc_b_b("main.set_squelch", theSquelchStatus);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_squelch_level - d:d - MAIN / Set Squelch Level, Return Old Squelch Level</summary>
		/// <param name="theSquelchLevel">Squelch Level</param>
		/// <returns>Old Squelch Level</returns>
		public Double Main_Set_Squelch_Level(
			Double theSquelchLevel
			)
		{
			return XmlRpc_d_d("main.set_squelch_level", theSquelchLevel);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.set_wf_sideband - n:s - MAIN / Set Waterfall Sideband to USB or LSB</summary>
		/// <param name="theWF_Sideband">Waterfall Sideband: USB or LSB</param>
		public void Main_Set_WF_Sideband(
			String theWF_Sideband
			)
		{
			XmlRpc_n_s("main.set_wf_sideband", theWF_Sideband);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.toggle_afc - b:n - MAIN / Toggle AFC Status</summary>
		/// <returns>New AFC Status</returns>
		public Boolean Main_Toggle_AFC()
		{
			return XmlRpc_b_n("main.toggle_afc");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.toggle_lock - b:n - MAIN / Toggle Transmit Lock Status, Return New Transmit Lock Status</summary>
		/// <returns>New Transmit Lock Status</returns>
		public Boolean Main_Toggle_Lock()
		{
			return XmlRpc_b_n("main.toggle_lock");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.toggle_reverse - b:n - MAIN / Toggle Reverse Sideband Status, Return New Reverse Sideband Status</summary>
		/// <returns>New Reverse Sideband Status</returns>
		public Boolean Main_Toggle_Reverse()
		{
			return XmlRpc_b_n("main.toggle_reverse");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.toggle_rsid - b:n - MAIN / Toggle RxID Status, Return New RxID Status</summary>
		/// <returns>New RxID Status</returns>
		public Boolean Main_Toggle_RSID()
		{
			return XmlRpc_b_n("main.toggle_rsid");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.toggle_squelch - b:n - MAIN / Toggle Squelch Status, Return New Toggle Status</summary>
		/// <returns>New Squelch Status</returns>
		public Boolean Main_Toggle_Squelch()
		{
			return XmlRpc_b_n("main.toggle_squelch");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.tune - n:n -  MAIN / Enter Tune Mode</summary>
		public void Main_Tune()
		{
			XmlRpc_n_n("main.tune");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>main.tx - n:n -  MAIN / Enter Transmit Mode</summary>
		public void Main_TX()
		{
			XmlRpc_n_n("main.tx");
		}
		//-----------------------------------------------------------------------------------------

		// MODEM FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>modem.get_afc_search_range - i:n - MODEM / Get Modem AFC Search Range</summary>
		/// <returns>Modem AFC Search Range</returns>
		public Int32 Modem_Get_AFC_Search_Range()
		{
			return XmlRpc_i_n("modem.get_afc_search_range");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_bandwidth - i:n - MODEM / Get Modem Bandwidth</summary>
		/// <returns>Modem Bandwidth</returns>
		public Int32 Modem_Get_Bandwidth()
		{
			return XmlRpc_i_n("modem.get_bandwidth");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_carrier - i:n - MODEM / Get Modem Carrier Frequency</summary>
		/// <returns>Modem Carrier Frequency</returns>
		public Int32 Modem_Get_Carrier()
		{
			return XmlRpc_i_n("modem.get_carrier");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_id - i:n - MODEM / Get Modem ID</summary>
		/// <returns>Modem ID</returns>
		public Int32 Modem_Get_ID()
		{
			return XmlRpc_i_n("modem.get_id");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_max_id - i:n - MODEM / Get Maximum Modem ID</summary>
		/// <returns>Maximum Modem ID</returns>
		public Int32 Modem_Get_Max_ID()
		{
			return XmlRpc_i_n("modem.get_max_id");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_name - i:n - MODEM / Get Modem Name (Digital Mode)</summary>
		/// <returns>Modem Name (Digital Mode)</returns>
		/// <remarks>Modem Name is the Digital Modulation Mode</remarks>
		public String Modem_Get_Name()
		{
			return XmlRpc_s_n("modem.get_name");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_names - A:n - MODEM / Get List of Modem Names</summary>
		/// <returns>List of Modem Names</returns>
		public String[] Modem_Get_Names()
		{
			XmlDocument myXmlResponse;			// Complete XML-RPC Response
			XmlNodeList myXmlNodeList;			// Data Value Node List
			String[] myStringArray;				// Returned Array

			//return XmlRpc_A_n("modem.get_names");
			myXmlResponse = XmlRpc_V_n("modem.get_names");

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				// Get List of Values in Array
				myXmlNodeList = myXmlResponse.SelectNodes("/methodResponse/params/param/value/array/data/value");
				int nNodes = myXmlNodeList.Count;
				if (nNodes > 0)
				{
					int i = 0;
					myStringArray = new String[nNodes];
					foreach (XmlNode myXmlNode in myXmlNodeList)
					{
						// Extract Value String from Each Value
						myStringArray[i++] = myXmlNode.InnerText;
					}
				}
				else
				{
					// Null Value if NO Array Elements
					myStringArray = null;
				}
			}
			else
			{
				myStringArray = null;
			}

			// Return Value
			return myStringArray;
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.get_quality - d:n -  MODEM / Get Signal Quality</summary>
		/// <returns>Signal Quality</returns>
		public Double Modem_Get_Quality()
		{
			return XmlRpc_d_n("modem.get_quality");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.inc_afc_search_range - i:i - MODEM / Increment AFC Search Range, Return New AFC Range</summary>
		/// <param name="theAFC_SearchRange">AFC Search Range Increment</param>
		/// <returns>New AFC Search Range</returns>
		public Int32 Modem_Inc_AFC_Search_Range(
			Int32 theAFC_SearchRange
			)
		{
			return XmlRpc_i_i("modem.inc_afc_search_range", theAFC_SearchRange);
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: Check Accuracy of Documentation for Signature: Modem_Inc_Bandwidth()
		/// <summary>modem.inc_bandwidth - i:i - MODEM / Increment Modem Bandwidth, Return New Modem Bandwidth</summary>
		/// <param name="theBandwidthIncrement"></param>
		/// <returns></returns>
		public Int32 Modem_Inc_Bandwidth(
			Int32 theBandwidthIncrement
			)
		{
			return XmlRpc_i_i("modem.inc_bandwidth", theBandwidthIncrement);
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: Check Accuracy of Documentation for Signature: Modem_Inc_Carrier()
		/// <summary>modem.inc_carrier - i:i - MODEM / Increment Modem Carrier Frequency, Return New Frequency</summary>
		/// <param name="theCarrierFrequencyIncrement">Modem Carrier Frequency Increment</param>
		/// <returns>New Modem Carrier Frequency</returns>
		public Int32 Modem_Inc_Carrier(
			Int32 theCarrierFrequencyIncrement
			)
		{
			return XmlRpc_i_i("modem.inc_carrier", theCarrierFrequencyIncrement);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.olivia.get_bandwidth - i:n - MODEM / Get Olivia Bandwidth</summary>
		/// <returns>Olivia Bandwidth</returns>
		public Int32 Modem_Olivia_Get_Bandwidth()
		{
			return XmlRpc_i_n("modem.olivia.get_bandwidth");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.olivia.get_tones - i:n -  MODEM / Get Olivia Tone Count</summary>
		/// <returns>Olivia Tone Count</returns>
		public Int32 Modem_Olivia_Get_Tones()
		{
			return XmlRpc_i_n("modem.olivia.get_tones");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.olivia.set_bandwidth - n:i - MODEM / Set Olivia Bandwidth</summary>
		/// <param name="theOliviaBandwidth">Olivia Bandwidth</param>
		public void Modem_Olivia_Set_Bandwidth(
			Int32 theOliviaBandwidth
			)
		{
			XmlRpc_n_i("modem.olivia.set_bandwidth", theOliviaBandwidth);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.olivia.set_tones - n:i - MODEM / Set Olivia Tone Count</summary>
		/// <param name="theOliviaTones">Olivia Tone Count</param>
		public void Modem_Olivia_Set_Tones(
			Int32 theOliviaTones
			)
		{
			XmlRpc_n_i("modem.olivia.set_tones", theOliviaTones);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.search_down - n:n - MODEM / Search Downward in Frequency</summary>
		public void Modem_Search_Down()
		{
			XmlRpc_n_n("modem.search_down");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.search_up - n:n - MODEM / Search Updward in Frequency</summary>
		public void Modem_Search_Up()
		{
			XmlRpc_n_n("modem.search_up");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.set_afc_search_range - n:i -  MODEM / Set AFC Search Range</summary>
		/// <param name="theModemSearchRange">AFC Search Range</param>
		public void Modem_Set_AFC_Search_Range(
			Int32 theModemSearchRange
			)
		{
			XmlRpc_n_i("modem.set_afc_search_range", theModemSearchRange);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.set_bandwidth - n:i - MODEM / Set Modem Bandwidth</summary>
		/// <param name="theBandwidth">Modem Bandwidth</param>
		public void Modem_Set_Bandwidth(
			Int32 theBandwidth
			)
		{
			XmlRpc_n_i("modem.set_bandwidth", theBandwidth);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.set_by_id - i:i -  MODEM / Set Current Modem By ID, Returns Old Modem ID</summary>
		/// <param name="theModemID">Modem ID</param>
		/// <returns>Old Modem ID</returns>
		public Int32 Modem_Set_By_ID(
			Int32 theModemID
			)
		{
			return XmlRpc_i_i("modem.set_by_id", theModemID);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.set_by_name - s:s - MODEM / Set Current Modem by Name, Return Old Modem Name</summary>
		/// <param name="theModemName">Modem Name</param>
		/// <returns>Old Modem Name</returns>
		public String Modem_Set_By_Name(
			String theModemName
			)
		{
			return XmlRpc_s_s("modem.set_by_name", theModemName);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>modem.set_carrier - i:i -  MODEM / Set Modem Carrier Frequency, Return Old Modem Carrier Frequency</summary>
		/// <param name="theCarrierFrequency">Modem Carrier Frequency</param>
		/// <returns>Old Modem Carrier Frequency</returns>
		public Int32 Modem_Set_Carrier(
			Int32 theCarrierFrequency
			)
		{
			return XmlRpc_i_i("modem.set_carrier", theCarrierFrequency);
		}
		//-----------------------------------------------------------------------------------------

		// NAVTEX FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>navtex.get_message - s:i - NAVTEX / Get Next Navtex / SitorB Message With Max Delay In Seconds, Empty String if Timeout</summary>
		/// <param name="theDelay">Max Delay In Seconds</param>
		/// <returns>Next Navtex / SitorB Message, Empty String if Timeout</returns>
		public String Navtex_Get_Message(
			Int32 theDelay
			)
		{
			return XmlRpc_s_i("navtex.get_message", theDelay);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>navtex.send_message - s:s NAVTEX / Send NavTex / SitorB Message, Return EMPTY if OK, Error Message if NOT</summary>
		/// <param name="theNavtexMessage">NavTex / SitorB Message</param>
		/// <returns>EMPTY if OK, Error Message if NOT</returns>
		public String Navtex_Send_Message(
			String theNavtexMessage
			)
		{
			return XmlRpc_s_s("navtex.send_message", theNavtexMessage);
		}
		//-----------------------------------------------------------------------------------------

		// RIG FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>rig.get_bandwidth - s:n - RIG / Get Rig Bandwidth Name</summary>
		/// <returns>Rig Bandwidth Name</returns>
		public String Rig_Get_Bandwidth()
		{
			return XmlRpc_s_n("rig.get_bandwidth");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.get_bandwidths - A:n - RIG / Get List of Rig Bandwidth Names</summary>
		/// <returns>List of Rig Bandwidth Names</returns>
		public String[] Rig_Get_Bandwidths()
		{
			//return XmlRpc_A_n("rig.get_bandwidths");

			XmlDocument myXmlResponse;			// Complete XML-RPC Response
			XmlNodeList myXmlNodeList;			// Data Value Node List
			String[] myStringArray;				// Returned Array

			//return XmlRpc_A_n("modem.get_names");
			myXmlResponse = XmlRpc_V_n("rig.get_bandwidths");

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				// Get List of Values in Array
				myXmlNodeList = myXmlResponse.SelectNodes("/methodResponse/params/param/value/array/data/value");
				int nNodes = myXmlNodeList.Count;
				if (nNodes > 0)
				{
					int i = 0;
					myStringArray = new String[nNodes];
					foreach (XmlNode myXmlNode in myXmlNodeList)
					{
						// Extract Value String from Each Value
						myStringArray[i++] = myXmlNode.InnerText;
					}
				}
				else
				{
					// Null Value if NO Array Elements
					myStringArray = null;
				}
			}
			else
			{
				myStringArray = null;
			}

			// Return Value
			return myStringArray;
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: Is this Really a Function, It is NOT on the FLDIGI Methods List?
		/// <summary>rig.get_frequency - d:n - RIG / Get Rig Frequency</summary>
		/// <returns>Rig Frequency</returns>
		public double Rig_Get_Frequency()
		{
			return XmlRpc_d_n("rig.get_frequency");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.get_mode - s:n - RIG / Get Mode</summary>
		/// <returns>Mode</returns>
		public String Rig_Get_Mode()
		{
			return XmlRpc_s_n("rig.get_mode");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.get_modes - A:n - RIG / Get List of Available Modes</summary>
		/// <returns>List of Available Modes</returns>
		public String[] Rig_Get_Modes()
		{
			//return XmlRpc_A_n("rig.get_modes");

			XmlDocument myXmlResponse;			// Complete XML-RPC Response
			XmlNodeList myXmlNodeList;			// Data Value Node List
			String[] myStringArray;				// Returned Array

			myXmlResponse = XmlRpc_V_n("rig.get_modes");

			// Decode Response
			if (m_ErrorCategory == FLDIGI_ErrorCategory.OK)
			{
				// Get List of Values in Array
				myXmlNodeList = myXmlResponse.SelectNodes("/methodResponse/params/param/value/array/data/value");
				int nNodes = myXmlNodeList.Count;
				if (nNodes > 0)
				{
					int i = 0;
					myStringArray = new String[nNodes];
					foreach (XmlNode myXmlNode in myXmlNodeList)
					{
						// Extract Value String from Each Value
						myStringArray[i++] = myXmlNode.InnerText;
					}
				}
				else
				{
					// Null Value if NO Array Elements
					myStringArray = null;
				}
			}
			else
			{
				myStringArray = null;
			}

			// Return Value
			return myStringArray;
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.get_name - s:n - RIG / Get Rig Name</summary>
		/// <returns>Rig Name</returns>
		public String Rig_Get_Name()
		{
			return XmlRpc_s_n("rig.get_name");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.get_notch - s:n -  RIG / Get Notch Filter Frequency Based On Waterfall Action</summary>
		/// <returns>Notch Filter Frequency</returns>
		/// <remarks>Based On Waterfall Action</remarks>
		public String Rig_Get_Notch()
		{
			return XmlRpc_s_n("rig.get_notch");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.release_control - n:n - RIG / Release XML-RPC Control</summary>
		/// <remarks>FLDIGI Will Return to CAT Control / Monitoring of RIG if this is Set Up</remarks>
		public void Rig_Release_Control()
		{
			XmlRpc_n_n("rig.release_control");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_bandwidth - n:s -  RIG / Select Bandwidth from Bandwidth List</summary>
		/// <param name="theBandwith">Bandwidth</param>
		/// <remarks>Bandwidth Should Be Previously Added In Bandwidth List</remarks>
		public void Rig_Set_Bandwidth(
			String theBandwith
			)
		{
			XmlRpc_n_s("rig.set_bandwidth", theBandwith);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_bandwidths - n:A - RIG / Set Bandwidth List</summary>
		/// <param name="theBandwithList">Bandwidth List</param>
		public void Rig_Set_Bandwidths(
			String[] theBandwithList
			)
		{
			//XmlRpc_n_A("rig.set_bandwidths", theBandwithList);

			XmlDocument myXmlMethodCall;
			XmlNode myArray;
			XmlNode myData;
			XmlNode myValue;
			XmlNode myString;

			// Create New Document
			myXmlMethodCall = new XmlDocument();

			// Create Array Node
			myArray = myXmlMethodCall.CreateElement("array");

			// Create and Append Data Node
			myData = myXmlMethodCall.CreateElement("data");
			myArray.AppendChild(myData);

			// Create a <data><string> Node for Each Array Element
			foreach (String myBandwidth in theBandwithList)
			{
				// Create and Append Value Node
				myValue = myXmlMethodCall.CreateElement("value");
				myData.AppendChild(myValue);

				// Create and Append String Node
				myString = myXmlMethodCall.CreateElement("string");
				myValue.AppendChild(myString);

				// Insert Text of String
				myString.InnerText = myBandwidth;
			}

			// Insert Value into Method Request and Query XML RPC Server
			XmlRpc_n_V("rig.set_bandwidths", myXmlMethodCall, myArray);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_frequency - d:d - RIG / Set Frequency, Return Old Frequency</summary>
		/// <param name="theFrequency">Frequency</param>
		/// <returns>Old Frequency</returns>
		public double Rig_Set_Frequency(
			Double theFrequency
			)
		{
			return XmlRpc_d_d("rig.set_frequency", theFrequency);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_mode - n:s - RIG / Select Mode from Mode List</summary>
		/// <param name="theArgument">Mode</param>
		/// <remarks>Mode Should Be Previously Added In Mode List</remarks>
		public void Rig_Set_Mode(
			String theArgument
			)
		{
			XmlRpc_n_s("rig.set_mode", theArgument);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_modes - n:A - RIG / Set Mode List</summary>
		/// <param name="theModeList">Mode List</param>
		public void Rig_Set_Modes(
			String[] theModeList
			)
		{
			XmlDocument myXmlMethodCall;
			XmlNode myArray;
			XmlNode myData;
			XmlNode myValue;
			XmlNode myString;

			//XmlRpc_n_A("rig.set_modes", theModeList);

			// Create New Document
			myXmlMethodCall = new XmlDocument();

			// Create Array Node
			myArray = myXmlMethodCall.CreateElement("array");

			// Create and Append Data Node
			myData = myXmlMethodCall.CreateElement("data");
			myArray.AppendChild(myData);

			// Create a <data><string> Node for Each Array Element
			foreach (String myMode in theModeList)
			{
				// Create and Append Value Node
				myValue = myXmlMethodCall.CreateElement("value");
				myData.AppendChild(myValue);

				// Create and Append String Node
				myString = myXmlMethodCall.CreateElement("string");
				myValue.AppendChild(myString);

				// Insert Text of String
				myString.InnerText = myMode;
			}

			// Insert Value into Method Request and Query XML RPC Server
			XmlRpc_n_V("rig.set_modes", myXmlMethodCall, myArray);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_name - n:s -  RIG / Set Rig Name</summary>
		/// <param name="theRigName">Rig Name</param>
		public void Rig_Set_Name(
			String theRigName
			)
		{
			XmlRpc_n_s("rig.set_name", theRigName);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.set_notch - n:i - RIG / Set Notch Frequency</summary>
		/// <param name="theNotch">Notch Frequency</param>
		/// <remarks>Notch Frequency is on Waterfall Display</remarks>
		public void Rig_Set_Notch(
			Int32 theNotch
			)
		{
			XmlRpc_n_i("rig.set_notch", theNotch);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rig.take_control - n:n RIG / Take XML-RPC Control</summary>
		/// <remarks>FLDIGI Will Relinquish CAT Control / Monitoring of RIG and Depend on Controlling Application</remarks>
		public void Rig_Take_Control()
		{
			XmlRpc_n_n("rig.take_control");
		}
		//-----------------------------------------------------------------------------------------

		// RXTX FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>rx.get_data - 6:n - Returns all RX data received since last query.</summary>
		/// <returns>RX data received since last query</returns>
		/// <remarks>Data is in BASE64 Format</remarks>
		public String RX_Get_Data()
		{
			return XmlRpc_6_n("rx.get_data");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>rxtx.get_data - 6:n - Returns all RXTX combined data since last query.</summary>
		/// <returns>RXTX combined data since last query</returns>
		/// <remarks>Data is in BASE64 Format</remarks>
		public String RXTX_Get_Data()
		{
			return XmlRpc_6_n("rxtx.get_data");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>tx.get_data - 6:n - TEXT / Returns ALL TX Data Transmitted Since Last Query</summary>
		/// <returns>TX data transmitted since last query</returns>
		public String TX_Get_Data()
		{
			return XmlRpc_6_n("tx.get_data");
		}
		//-----------------------------------------------------------------------------------------

		// SPOT FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>spot.get_auto - b:n - SPOT / Get AutoSpotter Status</summary>
		/// <returns>AutoSpotter Status</returns>
		public Boolean Spot_Get_Auto()
		{
			return XmlRpc_b_n("spot.get_auto");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>spot.pskrep.get_count - i:n - SPOT / Get Number of Callsigns Spotted in Current Session</summary>
		/// <returns>Number of Callsigns Spotted in Current Session</returns>
		public Int32 Spot_PSKrep_Get_Count()
		{
			return XmlRpc_i_n("spot.pskrep.get_count");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>spot.set_auto - n:b - SPOT / Set AutoSpotter Status, Return Old AutoSpotter Status</summary>
		/// <param name="theAutoStatus">AutoSpotter Status</param>
		/// <returns>Old AutoSpotter Status</returns>
		public Boolean Spot_Set_Auto(
			Boolean theAutoStatus
			)
		{
			return XmlRpc_b_b("spot.set_auto", theAutoStatus);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>spot.set_auto - n:b - SPOT / Toggle AutoSpotter Status, Return New AutoSpotter Status</summary>
		/// <returns>New AutoSpotter Status</returns>
		public Boolean Spot_Toggle_Auto()
		{
			return XmlRpc_b_n("spot.toggle_auto");
		}
		//-----------------------------------------------------------------------------------------

		// TEXT FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>text.add_tx - n:s - TEXT / Add String to TX Window</summary>
		/// <param name="theText">Text String</param>
		public void Text_Add_TX(
			String theText
			)
		{
			XmlRpc_n_s("text.add_tx", theText);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>text.add_tx_bytes - n:6 - TEXT / Add Text to TX Window in BASE64 Format</summary>
		/// <param name="theText">Text in BASE64 Format</param>
		public void Text_Add_TX_Bytes(
			String theText
			)
		{
			XmlRpc_n_6("text.add_tx_bytes", theText);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>text.clear_rx - n:n - TEXT / Clear RX Window</summary>
		public void Text_Clear_RX()
		{
			XmlRpc_n_n("text.clear_rx");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>text.clear_tx - n:n - TEXT / Clear TX Window</summary>
		public void Text_Clear_TX()
		{
			XmlRpc_n_n("text.clear_tx");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>text.get_rx - 6:ii -  TEXT / Get Substring of Characters from RX Window Starting at N, length L</summary>
		/// <param name="theStartIndex">Start Index</param>
		/// <param name="theLength">Substring Length</param>
		/// <returns>Text in BASE64 Format</returns>
		public String Text_Get_RX(
			Int32 theStartIndex,
			Int32 theLength
			)
		{
			return XmlRpc_6_ii("text.get_rx", theStartIndex, theLength);
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>TEXT / Get ALL Text from RX Window</summary>
		/// <returns>ALL Text from RX Window</returns>
		/// <remarks>This Combines text.get_rx_length and text.get_rx using 0 as Starting Index</remarks>
		public String Text_Get_RX_All()
		{
			Int32 myLength = Text_Get_RX_Length();	// text.get_rx_length
			return Text_Get_RX(0, myLength);		// text.get_rx
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>text.get_rx_length - i:n - TEXT / Get RX Window Character Count</summary>
		/// <returns>RX Window Character Count</returns>
		public Int32 Text_Get_RX_Length()
		{
			return XmlRpc_i_n("text.get_rx_length");
		}
		//-----------------------------------------------------------------------------------------

		// WEFAX FAMILY
		///////////////////////////////////////////////////////////////////////////////////////////

		/// TODO: Check Accuracy of Documentation for Signature: WEFAX_End_Reception()
		/// <summary>wefax.end_reception - s:n - WEFAX / End Reception</summary>
		/// <returns>WEFAX Image</returns>
		public String WEFAX_End_Reception()
		{
			return XmlRpc_s_n("wefax.end_reception");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>wefax.get_received_file - s:n - WEFAX / Get Received WEFAX File</summary>
		/// <returns>Received WEFAX File</returns>
		public String WEFAX_Get_Received_File()
		{
			return XmlRpc_s_n("wefax.get_received_file");
		}
		//-----------------------------------------------------------------------------------------

		/// <summary>wefax.state_string - s:n WEFAX / Get WEFAX Engine State (RX/TX) for Information</summary>
		/// <returns>WEFAX Engine State (RX/TX)</returns>
		public String WEFAX_Get_State_String()
		{
			return XmlRpc_s_n("wefax.state_string");
		}
		//-----------------------------------------------------------------------------------------
		/// TODO: Check Accuracy of Documentation for Signature: WEFAX_Send_File()
		/// <summary>wefax.send_file - s:si - WEFAX / Send WEFAX File</summary>
		/// <param name="theWEFAX_File">WEFAX File</param>
		/// <returns>Empty String if OK, Error Message if NOT</returns>
		public String WEFAX_Send_File(
			String theWEFAX_File
			)
		{
			return XmlRpc_s_s("wefax.send_file", theWEFAX_File);
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: What is the Return Value? WEFAX_Set_ADIF_Log()
		/// <summary>wefax.set_adif_log - s:b - WEFAX / Set ADIF Logging Status</summary>
		/// <param name="theADIF_LoggingStatus">ADIF Logging Status</param>
		/// <returns>?????</returns>
		public String WEFAX_Set_ADIF_Log(
			Boolean theADIF_LoggingStatus
			)
		{
			return XmlRpc_s_b("wefax.set_adif_log", theADIF_LoggingStatus);
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: What is the Return Value? WEFAX_Set_Max_Lines()
		/// <summary>wefax.set_max_lines - s:i - WEFAX / Set Maximum Lines</summary>
		/// <param name="theMaximumLines">Maximum Lines</param>
		/// <returns>?????</returns>
		public String WEFAX_Set_Max_Lines(
			Int32 theMaximumLines
			)
		{
			return XmlRpc_s_i("wefax.set_max_lines", theMaximumLines);
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: What is the Return Value? WEFAX_Set_TX_Abort_Flag()
		/// <summary>wefax.set_tx_abort_flag - s:n - WEFAX / Set TX Abort Flag - Cancel WEFAX Image Transmission</summary>
		/// <returns>?????</returns>
		/// <remarks>Cancel WEFAX Image Transmission</remarks>
		public String WEFAX_Set_TX_Abort_Flag()
		{
			return XmlRpc_s_n("wefax.set_tx_abort_flag");
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: What is the Return Value? WEFAX_Skip_APT()
		/// <summary>wefax.skip_apt - s:n - WEFAX / Skip APT During WEFAX Reception</summary>
		/// <returns>?????</returns>
		public String WEFAX_Skip_APT()
		{
			return XmlRpc_s_n("wefax.skip_apt");
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: What is the Return Value? WEFAX_Skip_Phasing()
		/// <summary>wefax.skip_phasing - s:n WEFAX / Skip Phasing During WEFAX Reception</summary>
		/// <returns>?????</returns>
		public String WEFAX_Skip_Phasing()
		{
			return XmlRpc_s_n("wefax.skip_phasing");
		}
		//-----------------------------------------------------------------------------------------

		/// TODO: What is the Return Value? WEFAX_Start_Manual_Reception()
		/// <summary>wefax.start_manual_reception - s:n WEFAX / Start FAX Image Reception in Manual Mode</summary>
		/// <returns>?????</returns>
		public String WEFAX_Start_Manual_Reception()
		{
			return XmlRpc_s_n("wefax.start_manual_reception");
		}
		//-----------------------------------------------------------------------------------------
	}
}
